from collections import namedtuple
import itertools
import re

from Bio import Seq, SeqRecord

from tools import general_tools


def first_overhang_seq(overhang, str_seq):
    """ confusing, but the first overhang will overlap with the first codon to
    some degree. Since the first breakpoint is at 0, AA2 for this breakpoint is
    at sequence position 0, while AA1 is at an imaginary sequence position -1.
    oh_pos represents the overhang position relative to AA1, so the number of
    overlapping positions is oh_in_seq=oh_pos+1. Therefore, the last oh_in_seq
    bases in the overhang and the first oh_in_seq bases in the sequence should
    be the same. It follows that we will need to add the 4-oh_in_seq leftmost
    overhang bases to the final fragment.
    """
    oh_pos, oh_seq = overhang
    if oh_pos == -1:
        return oh_seq  # just append oh onto sequence
    # number of bases on left side of seq that should be on right side of oh
    oh_in_seq = oh_pos + 1
    assert oh_seq[-oh_in_seq:] == str_seq[:oh_in_seq]
    return oh_seq[:4-oh_in_seq]


def last_overhang_seq(overhang, str_seq):
    """ confusing, but the last overhang will overlap with the last codon to
    some degree. Since the last breakpoint is at len(alignment), AA1 for this
    breakpoint is at the last sequence position, while AA2 is at an imaginary
    sequence position len(alignment). oh_pos represents the overhang position
    relative to AA1, so the number of overlapping positions is
    oh_in_seq=3-oh_pos. Therefore, the first oh_in_seq bases in the overhang
    and the last oh_in_seq bases in the sequence should be the same. It follows
    that we will need to add the 4-oh_in_seq rightmost overhang bases to the
    final fragment.
    """
    oh_pos, oh_seq = overhang
    if oh_pos == 3:
        return oh_seq  # just append oh onto sequence
    # number of bases on right side of seq that should be in left side of oh
    oh_in_seq = 3 - oh_pos
    assert oh_seq[:oh_in_seq] == str_seq[-oh_in_seq:]
    return oh_seq[oh_in_seq-4:]


def _get_valid_codon(AA, pattern):
    residue_cdns = general_tools.rev_code[AA]
    possible_cdns = [c for c in residue_cdns if re.fullmatch(pattern, c)]
    cdn = possible_cdns[0]
    return cdn


def overhang_CDN_seq(overhang, cdn1, cdn2, block_front_back):
    position, oh_seq = overhang

    # build codon patterns for both codons
    overall_pattern = '.' * position + oh_seq + '.' * (2 - position)
    pattern1 = overall_pattern[:3]
    pattern2 = overall_pattern[3:]

    AA1 = general_tools.code[cdn1]
    AA2 = general_tools.code[cdn2]

    cdn1 = _get_valid_codon(AA1, pattern1)
    cdn2 = _get_valid_codon(AA2, pattern2)

    cdn_seq = cdn1 + cdn2

    assert block_front_back in ('front', 'back')
    if block_front_back == 'front':
        return cdn_seq[position:]

    if position == 2:
        return cdn_seq
    return cdn_seq[:-2 + position]


# Script for confirming order against amino acid sequence below

class Restriction_Enzyme(object):
    def __init__(self, rec_site, offset, cut_len, sticky_end_strand):
        self.site = rec_site
        self.offset = offset
        self.cut_len = cut_len

        # which strand sticky end stays on relative to site, 'same' or 'rc'
        assert sticky_end_strand == 'same' or sticky_end_strand == 'rc'
        self.sticky_end_strand = sticky_end_strand


bsaI = Restriction_Enzyme('GGTCTC', 7, 4, 'rc')

Overhang_set = namedtuple('Overhang_set', ['fwd_start', 'fwd_end', 'rev_start',
                                           'rev_end'])


def same_strand_cut(seq, rs_site, enzyme):
    se_len = enzyme.cut_len  # length of sticky ends
    offset = enzyme.offset
    if enzyme.sticky_end_strand == 'same':
        cut_site = rs_site + offset + se_len
        new_seq = seq[cut_site:]
    else:
        cut_site = rs_site + offset
        cut_seq = seq[cut_site:]
        new_seq = cut_seq[:se_len].lower() + cut_seq[se_len:]
    return new_seq


def rc_strand_cut(seq, rs_site, enzyme):
    # assume same and rc srands end at same position
    se_len = enzyme.cut_len  # length of sticky ends
    offset = enzyme.offset
    if enzyme.sticky_end_strand == 'same':
        cut_site = -rs_site - offset
        cut_seq = seq[:cut_site]
        new_seq = cut_seq[:-se_len] + cut_seq[-se_len:].lower()
    else:
        cut_site = -rs_site - offset - se_len
        new_seq = seq[:cut_site]
    return new_seq


def reverse_complements(oh1, oh2):
    s1 = Seq.Seq(oh1)
    s2 = Seq.Seq(oh2)
    return s1 == s2.reverse_complement()


def concatenate_frags(frag1, frag2):
    # order matters, result is frag1 + frag2
    frag1_overhangs = frag1.get_overhangs()
    frag2_overhangs = frag2.get_overhangs()
    assert frag1_overhangs.fwd_end or frag2_overhangs.rev_start
    if frag1_overhangs.fwd_end:
        frag1_oh = frag1_overhangs.fwd_end
        frag2_oh = frag2_overhangs.rev_end
        assert reverse_complements(frag1_oh, frag2_oh)
        oh_len = len(frag1_oh)
        fwd_seq = frag1.fwd[:-oh_len] + frag1_oh.upper() + frag2.fwd
        rev_seq = frag2.rev[:-oh_len] + frag2_oh.upper() + frag1.rev
    else:
        frag1_oh = frag1_overhangs.rev_start
        frag2_oh = frag2_overhangs.fwd_start
        assert reverse_complements(frag1_oh, frag2_oh)
        oh_len = len(frag1_oh)
        fwd_seq = frag1.fwd + frag2_oh.upper() + frag2.fwd[oh_len:]
        rev_seq = frag2.rev + frag1_oh.upper() + frag1.rev[oh_len:]

    return Fragment(fwd_seq, rev_seq)


def get_seq_overhangs(seq):
    start_se_len = re.search('[A-Z]', seq).start()  # fwd seq beginning
    end_se_len = re.search('[A-Z]', seq[::-1]).start()  # fwd seq end
    if end_se_len:
        return seq[:start_se_len], seq[-end_se_len:]
    return seq[:start_se_len], ''


class Fragment(object):
    def __init__(self, fwd_seq: str, rev_seq: str):
        self.fwd = fwd_seq
        self.rev = rev_seq

    @classmethod
    def from_SeqRecord(cls, sr: SeqRecord.SeqRecord) -> 'Fragment':
        fwd = str(sr.seq).upper()
        rev = str(sr.seq.reverse_complement()).upper()
        return cls(fwd, rev)

    def digest(self, enzyme: Restriction_Enzyme):
        # find restriction sites
        fwd_sites = [m.start() for m in re.finditer('(?=%s)' % enzyme.site,
                                                    self.fwd)]
        if len(fwd_sites) > 1:
            raise Exception('More than one foward restriction site found.')
        fwd_site_index = fwd_sites[0]
        rev_sites = [m.start() for m in re.finditer('(?=%s)' % enzyme.site,
                                                    self.rev)]
        if len(rev_sites) > 1:
            raise Exception('More than one reverse restriction site found.')
        rev_site_index = rev_sites[0]

        # create new forward seq
        new_fwd = self.fwd
        new_fwd = same_strand_cut(new_fwd, fwd_site_index, enzyme)
        new_fwd = rc_strand_cut(new_fwd, rev_site_index, enzyme)

        # create new reverse seq
        new_rev = self.rev
        new_rev = same_strand_cut(new_rev, rev_site_index, enzyme)
        new_rev = rc_strand_cut(new_rev, fwd_site_index, enzyme)

        return Fragment(new_fwd, new_rev)

    def get_overhangs(self):
        fwd_start_oh, fwd_end_oh = get_seq_overhangs(self.fwd)
        rev_start_oh, rev_end_oh = get_seq_overhangs(self.rev)
        return Overhang_set(fwd_start_oh, fwd_end_oh, rev_start_oh, rev_end_oh)

    def translate(self):
        dna = self.fwd.upper()
        start_index = re.search('ATG', dna).start()
        seq = Seq.Seq(dna[start_index:])
        return str(seq.translate())

    def __repr__(self):
        """ Temporary function for testing, works with bsaI specifically.
        """
        rev = self.rev[::-1]
        if self.fwd[:4].islower():
            return f'fwd: {self.fwd}\nrev:     {rev}'
        return f'fwd: {self.fwd}\nrev: {rev}'


def check_seq(frags):
    CDN_frags, AA_frags = zip(*frags)
    AA_seq = ''.join(AA_frags)
    building_frag = CDN_frags[0].digest(bsaI)
    for next_frag in CDN_frags[1:]:
        next_frag = next_frag.digest(bsaI)
        building_frag = concatenate_frags(building_frag, next_frag)
    CDN_seq = building_frag.translate()
    assert AA_seq == CDN_seq or (CDN_seq[0] == 'M' and AA_seq == CDN_seq[1:])
    assert not [m.start() for m in re.finditer('GGTCTC', building_frag.fwd)]
    assert not [m.start() for m in re.finditer('GGTCTC', building_frag.rev)]


def parent_frag_tuple(sr_id):
    """ Return a tuple (parent name, frag name) from the sr_id. Needs to
    handle case where parent name has underscores."""
    *name_list, frag = sr_id.split('_')
    name_list = [x + '_' for x in name_list[:-1]] + [name_list[-1]]
    parent_name = ''.join(name_list)
    return parent_name, frag


def verify_fragments(order_srs, AA_srs, lib_bps):
    order_frags = {parent_frag_tuple(sr.id): Fragment.from_SeqRecord(sr)
                   for sr in order_srs}
    order_frags = {(p_name, int(blk[-1])-1): v for (p_name, blk), v
                   in order_frags.items()}
    block_dict = {i: (lib_bps[i], b) for i, b in enumerate(lib_bps[1:])}
    AA_frags = {}
    for sr in AA_srs:
        seq = str(sr.seq)
        for b, (start, end) in block_dict.items():
            AA_frags.update({(sr.id, b): seq[start: end].replace('-', '')})

    combined_frags = {}
    for k, v in order_frags.items():
        combined_frags.update({k: (v, AA_frags[k])})

    pools = {}
    for (parent, block), v in combined_frags.items():
        if block in pools:
            pools[block].append((parent, v))
        else:
            pools[block] = [(parent, v)]
    pools = [pools[i] for i in pools]

    total_chimeras = len(pools[0]) ** len(pools)
    progress = general_tools.ProgressOutput(total_chimeras)
    for j, combo in enumerate(itertools.product(*pools)):
        progress.update(j)
        seq_frags = [i[1] for i in combo]
        check_seq(seq_frags)
    print('all sequences good')



""" Generate DNA fragments from chosen library for ordering.

Run syntax: "python step4.py <AA_alignment_fn> <CDN_alignment_fn>
    <chosen_lib_fn> <frags_order_fn>"
Example: "python step4.py bgl3_AA.fasta bgl3_CDN.fasta bgl3_chosen_lib.json"

Command Line Args:
    AA_alignment_fn: name of aligned amino acid sequences file, from step 1
    CDN_alignment_fn: name of aligned codon sequences file, from step1
    chosen_lib_fn: name of chosen library file, from step 3
    frags_order_fn: name of frags order fasta file

Output:
    DNA fragments for ordering in <frags_order_fn>

You can also replace all instances of "sys.argv" in the code with the input
filenames directly, then run "python step1.py".
"""

import json
import sys

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq

from tools import step4_tools

AA_alignment_fn = sys.argv[1]  # aligned AA sequences
CDN_alignment_fn = sys.argv[2]  # aligned codon sequences
chosen_lib_fn = sys.argv[3]  # chosen library, (lib_bps, lib_attrs)
frags_order_fn = sys.argv[4]  # fragment order output

# load amino acid alignment
alignment_SRs = list(SeqIO.parse(AA_alignment_fn, 'fasta'))
AA_names = [str(i.id) for i in alignment_SRs]
AA_seqs = [str(i.seq) for i in alignment_SRs]
AA_alignment = list(zip(*AA_seqs))  # list of AAs at each position

# load codon sequences
DNA_SRs = list(SeqIO.parse(CDN_alignment_fn, 'fasta'))
DNA_names = [str(i.id) for i in DNA_SRs]
assert DNA_names == AA_names
CDN_seqs = []
for sr in DNA_SRs:
    dna_seq = str(sr.seq)
    cdns = [dna_seq[index: index+3] for index in range(0, len(dna_seq), 3)]
    name = str(sr.id)
    CDN_seqs.append((name, cdns))

# load library breakpoints and overhangs
with open(chosen_lib_fn, 'r') as f:
    breakpoints, lib_attrs = json.load(f)
overhangs = lib_attrs['GG_sites']

# list of fragment to order
frags = []

# bsaI recognition sites with throwaway bases front and back
bsaI_start = 'ctagc' + 'ggtctc' + 'c'
bsaI_end = 'c' + 'gagacc' + 'gactc'

# first and last breakpoints are special -> different procedures for first and
# last blocks

print('Constructing fragments', flush=True)

# starting block
for p_name, cdn_seq in CDN_seqs:
    bp2 = breakpoints[1]  # index of breakpoint at end of block
    # make fragment sequence with breakpoint residues stripped off
    str_seq = ''.join(cdn_seq[:bp2-1]).replace('-', '')

    # sequence that gets added to front of fragment
    overhang_seq1 = step4_tools.first_overhang_seq(overhangs[0], str_seq)
    front_addition = bsaI_start + overhang_seq1

    # sequence that gets added to back of fragment
    bp2_cdn1, bp2_cdn2 = cdn_seq[bp2-1], cdn_seq[bp2]
    overhang_seq2 = step4_tools.overhang_CDN_seq(overhangs[1], bp2_cdn1,
                                                 bp2_cdn2,
                                                 block_front_back='back')
    back_addition = overhang_seq2 + bsaI_end

    frag_seq = front_addition + str_seq + back_addition
    frag_sr = SeqRecord(Seq(frag_seq), p_name + '_frag1')
    frags.append(frag_sr)

# middle blocks
# breakpoint indices and overhang information for start and end of blocks
front_bps, front_ohs = breakpoints[1:-2], overhangs[1:-2]  # starts of blocks
back_bps, back_ohs = breakpoints[2:-1], overhangs[2:-1]  # ends of blocks
blocks_iter = zip(front_bps, front_ohs, back_bps, back_ohs)
for block_num, (bp1, oh1, bp2, oh2) in enumerate(blocks_iter, 2):
    for p_name, cdn_seq in CDN_seqs:
        # make fragment sequence with breakpoint residues stripped off
        str_seq = ''.join(cdn_seq[bp1+1:bp2-1]).replace('-', '')

        # sequence that gets added to front of fragment
        bp1_cdn1, bp1_cdn2 = cdn_seq[bp1-1], cdn_seq[bp1]
        oh1_seq = step4_tools.overhang_CDN_seq(oh1, bp1_cdn1, bp1_cdn2,
                                               block_front_back='front')
        front_addition = bsaI_start + oh1_seq

        # sequence that gets added to back of fragment
        bp2_cdn1, bp2_cdn2 = cdn_seq[bp2-1], cdn_seq[bp2]
        oh2_seq = step4_tools.overhang_CDN_seq(oh2, bp2_cdn1, bp2_cdn2,
                                               block_front_back='back')
        back_addition = oh2_seq + bsaI_end

        frag_seq = front_addition + str_seq + back_addition
        frag_sr = SeqRecord(Seq(frag_seq), p_name + f'_frag{block_num}')
        frags.append(frag_sr)

# ending block
block_num += 1
for p_name, cdn_seq in CDN_seqs:
    bp1 = breakpoints[-2]  # index of breakpoint at beginning of block
    # make fragment sequence with breakpoint residues stripped off
    str_seq = ''.join(cdn_seq[bp1+1:]).replace('-', '')

    # sequence that gets added to front of fragment
    bp1_cdn1, bp1_cdn2 = cdn_seq[bp1-1], cdn_seq[bp1]
    oh1_seq = step4_tools.overhang_CDN_seq(overhangs[-2], bp1_cdn1, bp1_cdn2,
                                           block_front_back='front')
    front_addition = bsaI_start + oh1_seq

    overhang_seq = step4_tools.last_overhang_seq(overhangs[-1], str_seq)
    back_addition = overhang_seq + bsaI_end

    # sequence that gets added to back of fragment
    frag_seq = front_addition + str_seq + back_addition
    frag_sr = SeqRecord(Seq(frag_seq), p_name + f'_frag{block_num}')
    frags.append(frag_sr)

SeqIO.write(frags, frags_order_fn, 'fasta')

print('Verifying fragment order against amino acid sequences', flush=True)
step4_tools.verify_fragments(frags, alignment_SRs,  breakpoints)
